#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/atomic.h>
#if defined( CONFIG_USART )
#define BAUD CONFIG_USART_BAUD
#include <avr/sleep.h>
#include <util/setbaud.h>
#endif

#include <stdio.h>

#include "../include/uCTest/uCTest.h"
#include "System.h"
#include "commands.h"

#ifndef CONFIG_USART
#define SOR ( *( ( volatile char * ) 0x20 ) )
#define SEXITR ( *( ( volatile char * ) 0x21 ) )
#endif

static int uart_putchar( char c, FILE *stream );
static FILE mystdout = FDEV_SETUP_STREAM( uart_putchar, NULL, _FDEV_SETUP_WRITE );

FILE *test_out;

static inline void write_char( char c ) {
#ifndef CONFIG_USART
    SOR = c;
#else
    if( c == '\n' )
        write_char( '\r' );
    loop_until_bit_is_set( UCSR0A, UDRE0 );
    UDR0 = c;
#endif
}

static int uart_putchar( char c, FILE *stream ) {
    ( ( void ) stream );

#if defined( CONFIG_OUTPUT_JSON )
    extern int uctest_shouldEscape( char c );
    if( uctest_shouldEscape( c ) ) {
        switch( c ) {
            case '\n':
                write_char( '\\' );
                write_char( 'n' );
                return 0;
            default: write_char( '\\' ); break;
        }
    }
#endif

    write_char( c );
    return 0;
}

void uctest_system_init_hardware( void ) {
#ifdef CONFIG_USART
    UBRR0 = UBRR_VALUE;
#if USE_2X
    UCSR0A |= ( 1 << U2X0 );
#else
    UCSR0A &= ~( 1 << U2X0 );
#endif
    UCSR0B = ( 1 << TXEN0 );
#endif
    test_out = &mystdout;

#if defined( CONFIG_COUNT_CYCLES ) && !defined( CONFIG_OUTPUT_SIMULATOR )
    ATOMIC_BLOCK( ATOMIC_FORCEON ) {
        TCCR1B = ( 1 << CS10 );
        TIMSK1 = ( 1 << TOIE1 );
    }
#endif
}

#ifdef CONFIG_COUNT_CYCLES
#ifndef CONFIG_OUTPUT_SIMULATOR
static volatile uint16_t ovf;
static volatile uint16_t clk_start;

ISR( TIMER1_OVF_vect ) {
    ovf++;
}
#endif

void uctest_clkMeasureStart( void ) {
#ifndef CONFIG_OUTPUT_SIMULATOR
    ATOMIC_BLOCK( ATOMIC_RESTORESTATE ) {
        ovf = 0;
        clk_start = TCNT1;
    }
#else
    putc( SIM_CMD_CLK_START, test_out );
#endif
}

void uctest_clkMeasureStop( unsigned long *clk ) {
#ifndef CONFIG_OUTPUT_SIMULATOR
    uint16_t clk_end;
    uint16_t ovf_end;
    ATOMIC_BLOCK( ATOMIC_RESTORESTATE ) {
        clk_end = TCNT1;
        ovf_end = ovf;
    }
    *clk += ( ( ovf_end * ( ( ( unsigned long ) UINT16_MAX ) + 1 ) ) + ( clk_end - clk_start ) );
#else
    ( ( void ) clk );
    putc( SIM_CMD_CLK_END, test_out );
#endif
}
#endif

void exit( int status ) {
    __builtin_avr_cli();
#ifndef CONFIG_USART
    SEXITR = status;
#else
    ( ( void ) status );
    set_sleep_mode( SLEEP_MODE_PWR_DOWN );
    sleep_enable();
    sleep_bod_disable();
    sleep_cpu();
#endif
    while( 1 )
        ;
}
