/**
 * @ingroup    uCTest
 * @{
 */

#ifdef __cplusplus
extern "C" {
#endif

#ifndef _UCTEST_COMMANDS_H_
#if !defined( __DOXYGEN__ )
#define _UCTEST_COMMANDS_H_
#endif
/**
 * @defgroup cmds Commands: Commands for the simulator
 *
 * These commands are used for the communication between the simulator and the test framework on the
 * controller.
 *
 * @{
 */
/**
 * @brief      Starting the testframework
 */
#define SIM_CMD_START 0x80
/**
 * @brief      Start the global setup routine
 */
#define SIM_CMD_GLOBAL_SETUP 0x81
/**
 * @brief      Start the global teardown routine
 */
#define SIM_CMD_GLOBAL_TEARDOWN 0x82
/**
 * @brief      Register a new test
 *
 * This command register a new test. After the control byte the address on the flash is transmitted.
 * For the AVR devices the address is 16 bit long.
 */
#define SIM_CMD_TEST_REGISTER 0x83
/**
 * @brief      The testframework has ended
 */
#define SIM_CMD_END 0x84

/**
 * @brief      Start a suite
 *
 * Indicates that a new suite is running. After the command byte the number of tests (16 bit) are
 * transmitted and than the flash address (16 bit) where the name of the suite is located.
 */
#define SIM_CMD_SUITE_START 0x85
/**
 * @brief       End the suite
 */
#define SIM_CMD_SUITE_END 0x86

/**
 * @brief      Start a test
 *
 * A test has started. After the command byte the next two bytes are the flash address of the name
 * of the test.
 */
#define SIM_CMD_TEST_START 0x8A
/**
 * @brief      The test finished sucessful
 */
#define SIM_CMD_TEST_SUCCESS 0x8B
/**
 * @brief      The test finished with a failure
 */
#define SIM_CMD_TEST_FAILURE 0x8C
/**
 * @brief      Print a user message
 *
 * Prints a user message. After the command byte a string is send. This string is terminated by a
 * zero byte.
 */
#define SIM_CMD_TEST_MSG 0x8D
/**
 * @brief      Print a failure message
 *
 * Prints a failure message. After the command byte a string is send. This string is terminated by a
 * zero byte.
 */
#define SIM_CMD_TEST_MSG_FAILURE 0x8E

/**
 * @brief       Start a time measurement
 */
#define SIM_CMD_CLK_START 0x90
/**
 * @brief       Stop the time measurement
 */
#define SIM_CMD_CLK_END 0x91
/** @} */

#endif /** _UCTEST_COMMANDS_H_ */

#ifdef __cplusplus
}
#endif
/** @} */
