#include <stdarg.h>
#include <stdio.h>

#include "../include/uCTest/uCTest.h"

#define XSTR( X ) UCTEST_XSTR( X )

extern FILE *test_out;

#if defined( CONFIG_OUTPUT_NONE )
const __flash char uctest_eqFailMsg[] = "";

const __flash char uctest_nearFailMsg[] = "";

const __flash char uctest_simpleFailMsg[] = "";

const __flash char uctest_deathFailMsg[] = "";

void uctest_printRuntimeStats( int suites, int tests ) {
    ( void ) suites;
    ( void ) tests;
}

void uctest_printRuntimeStatsEnd( int suites, int tests, unsigned long cycles ) {
    ( void ) suites;
    ( void ) tests;
    ( void ) cycles;
}

void uctest_printStartup( int suites, int tests ) {
    ( void ) suites;
    ( void ) tests;
}

void uctest_printGlobalSetup( void ) {
}

void uctest_printGlobalTeardown( void ) {
}

void uctest_printSuiteStats( const __flash char *suite, int tests ) {
    ( void ) suite;
    ( void ) tests;
}

void uctest_printSuiteStatsBegin( const __flash char *suite, int tests ) {
    ( void ) suite;
    ( void ) tests;
}

void uctest_printSuiteStatsEnd( const __flash char *suite, int tests, int failed,
                                unsigned long cycles ) {
    ( void ) suite;
    ( void ) tests;
    ( void ) failed;
    ( void ) cycles;
}

void uctest_printTest( const __flash char *mode, const __flash struct uctest_test *test ) {
    ( void ) mode;
    ( void ) test;
}

void uctest_printTestBegin( const __flash struct uctest_test *test ) {
    ( void ) test;
}

void uctest_printTestSuccess( const __flash struct uctest_test *test, unsigned long cycles ) {
    ( void ) test;
    ( void ) cycles;
}

void uctest_printAllTestsEnd( int testsSuccess, int testsFailure ) {
    ( void ) testsSuccess;
    ( void ) testsFailure;
}

void uctest_printFailedTestsAtEnd( int tests ) {
    ( void ) tests;
}

void uctest_printTestFailure( const __flash struct uctest_test *test, unsigned long cycles ) {
    ( void ) test;
    ( void ) cycles;
}

void uctest_printFailureLocation( const char *file, int line ) {
    ( void ) file;
    ( void ) line;
}

void uctest_printUserMessage( const __flash char *fmt, va_list args ) {
    ( void ) fmt;
    ( void ) args;
}
#elif defined( CONFIG_OUTPUT_SIMULATOR )
#include "commands.h"

void uctest_printRuntimeStats( int suites, int tests ) {
    ( void ) suites;
    ( void ) tests;
}

void uctest_printRuntimeStatsEnd( int suites, int tests, unsigned long cycles ) {
    putc( SIM_CMD_END, test_out );
    ( void ) suites;
    ( void ) tests;
    ( void ) cycles;
}

void uctest_printStartup( int suites, int tests ) {
    ( void ) suites;
    ( void ) tests;
    putc( SIM_CMD_START, test_out );
}

void uctest_printGlobalSetup( void ) {
    putc( SIM_CMD_GLOBAL_SETUP, test_out );
}

void uctest_printGlobalTeardown( void ) {
    putc( SIM_CMD_GLOBAL_TEARDOWN, test_out );
}

void uctest_printTestRegister( const __flash struct uctest_test *test ) {
    putc( SIM_CMD_TEST_REGISTER, test_out );
    putc( ( ( int ) test ) >> 8, test_out );
    putc( ( ( int ) test ) & 0xFF, test_out );
}

void uctest_printSuiteStats( const __flash char *suite, int tests ) {
    ( void ) suite;
    ( void ) tests;
}

void uctest_printSuiteStatsBegin( const __flash char *suite, int tests ) {
    putc( SIM_CMD_SUITE_START, test_out );
    putc( tests >> 8, test_out );
    putc( tests & 0xFF, test_out );
    putc( ( ( uint16_t ) suite ) >> 8, test_out );
    putc( ( ( uint16_t ) suite ) & 0xFF, test_out );
}

void uctest_printSuiteStatsEnd( const __flash char *suite, int tests, int failed,
                                unsigned long cycles ) {
    ( void ) suite;
    ( void ) tests;
    ( void ) failed;
    ( void ) cycles;
    putc( SIM_CMD_SUITE_END, test_out );
}

void uctest_printTest( const __flash char *mode, const __flash struct uctest_test *test ) {
    ( void ) mode;
    ( void ) test;
}

void uctest_printTestBegin( const __flash struct uctest_test *test ) {
    putc( SIM_CMD_TEST_START, test_out );
    putc( ( ( uint16_t ) test->test_name ) >> 8, test_out );
    putc( ( ( uint16_t ) test->test_name ) & 0xFF, test_out );
}

void uctest_printTestSuccess( const __flash struct uctest_test *test, unsigned long cycles ) {
    putc( SIM_CMD_TEST_SUCCESS, test_out );
    ( void ) test;
    ( void ) cycles;
}

void uctest_printAllTestsEnd( int testsSuccess, int testsFailure ) {
    ( void ) testsSuccess;
    ( void ) testsFailure;
}

void uctest_printFailedTestsAtEnd( int tests ) {
    ( void ) tests;
}

void uctest_printTestFailure( const __flash struct uctest_test *test, unsigned long cycles ) {
    putc( SIM_CMD_TEST_FAILURE, test_out );
    ( void ) test;
    ( void ) cycles;
}

void uctest_printFailureLocation( const char *file, int line ) {
    putc( SIM_CMD_TEST_MSG_FAILURE, test_out );
    fprintf_P( test_out, XSTR( "%s:%d" ), file, line );
    putc( 0x00, test_out );
}

void uctest_printUserMessage( const __flash char *fmt, va_list args ) {
    putc( SIM_CMD_TEST_MSG, test_out );
    vfprintf_P( test_out, fmt, args );
    putc( 0x00, test_out );
}
#elif defined( CONFIG_OUTPUT_JSON )
void uctest_printRuntimeStats( int suites, int tests ) {
    ( void ) suites;
    ( void ) tests;
}

void uctest_printRuntimeStatsEnd( int suites, int tests, unsigned long cycles ) {
    fprintf_P( test_out, XSTR( "    {}\n  ],\n  \"time\": %lu,\n" ), cycles );
    ( void ) suites;
    ( void ) tests;
}

void uctest_printStartup( int suites, int tests ) {
    ( void ) suites;
    ( void ) tests;
    fprintf_P( test_out, XSTR( "{\n  \"name\": \"AllTests\",\n  \"testsuites\": [\n" ), tests );
}

void uctest_printGlobalSetup( void ) {
}

void uctest_printGlobalTeardown( void ) {
}

void uctest_printTestRegister( const __flash struct uctest_test *test ) {
    ( void ) test;
}

void uctest_printSuiteStats( const __flash char *suite, int tests ) {
    ( void ) suite;
    ( void ) tests;
}

void uctest_printSuiteStatsBegin( const __flash char *suite, int tests ) {
#ifdef __AVR__
    fprintf_P( test_out, XSTR( "    {\n      \"name\": \"%S\",\n      \"testsuite\": [\n" ), suite,
               tests );
#else
    fprintf_P( test_out, XSTR( "    {\n      \"name\": \"%s\",\n      \"testsuite\": [\n" ), suite,
               tests );
#endif
}

void uctest_printSuiteStatsEnd( const __flash char *suite, int tests, int failed,
                                unsigned long cycles ) {
    fprintf_P(
        test_out,
        XSTR(
            "        {}\n      ],\n      \"time\": %lu,\n      \"tests\": %d,\n      \"failures\": "
            "%d\n    },\n" ),
        cycles, tests, failed );
    ( void ) suite;
}

void uctest_printTest( const __flash char *mode, const __flash struct uctest_test *test ) {
    ( void ) mode;
    ( void ) test;
}

static int test_failed;
void uctest_printTestBegin( const __flash struct uctest_test *test ) {
    test_failed = 0;
#ifdef __AVR__
    fprintf_P(
        test_out,
        XSTR( "        {\n          \"classname\": \"%S\",\n          \"name\": \"%S\",\n    "
              "      \"status\": \"RUN\",\n" ),
        test->suite_name, test->test_name );
#else
    fprintf_P(
        test_out,
        XSTR( "        {\n          \"classname\": \"%s\",\n          \"name\": \"%s\",\n    "
              "      \"status\": \"RUN\",\n" ),
        test->suite_name, test->test_name );
#endif
}

void uctest_printTestSuccess( const __flash struct uctest_test *test, unsigned long cycles ) {
    fprintf_P(
        test_out,
        XSTR( "          \"result\": \"COMPLETED\",\n          \"time\": %lu\n        },\n" ),
        cycles );
    ( void ) test;
}

void uctest_printTestFailure( const __flash struct uctest_test *test, unsigned long cycles ) {
    if( cycles ) {
        fprintf_P( test_out, XSTR( "            {}\n          ],\n" ) );
        uctest_printTestSuccess( test, cycles );
    }
}

void uctest_printAllTestsEnd( int testsSuccess, int testsFailure ) {
    fprintf_P( test_out, XSTR( "  \"tests\": %d,\n  \"failures\": %d\n}\n" ),
               ( testsSuccess + testsFailure ), testsFailure );
}

void uctest_printFailedTestsAtEnd( int tests ) {
    ( void ) tests;
}

static int inString = 0;
void uctest_printFailureLocation( const char *file, int line ) {
    if( !test_failed ) {
        test_failed = !test_failed;
        fprintf_P( test_out, XSTR( "          \"failures\": [\n" ) );
    }

    fprintf_P( test_out, XSTR( "            {\n              \"failure\": \"" ), file, line );
    inString = 1;
    fprintf_P( test_out, XSTR( "%s:%d\n" ), file, line );
}

void uctest_printUserMessage( const __flash char *fmt, va_list args ) {
    vfprintf_P( test_out, fmt, args );
    inString = 0;

    fprintf_P( test_out, XSTR( "\",\n              \"type\": \"\"\n            },\n" ) );
}

int uctest_shouldEscape( char c ) {
    switch( c ) {
        case '\"':
        case '\'':
        case '\\':
        case '\n':
        case '/':
        case '\t': return inString;
        default: return 0;
    }
}
#else
void uctest_printRuntimeStats( int suites, int tests ) {
    fprintf_P( test_out, XSTR( "\e[32m[==========]\e[0m Running %d tests from %d test suites.\n" ),
               tests, suites );
}

void uctest_printRuntimeStatsEnd( int suites, int tests, unsigned long cycles ) {
    fprintf_P(
        test_out,
        XSTR( "\e[32m[==========]\e[0m %d tests from %d test suites ran. (%lu cycles total)\n" ),
        tests, suites, cycles );
}

void uctest_printStartup( int suites, int tests ) {
    fprintf_P( test_out, XSTR( "Running main\n" ) );
    uctest_printRuntimeStats( suites, tests );
}

void uctest_printGlobalSetup( void ) {
    fprintf_P( test_out, XSTR( "\e[32m[----------]\e[0m Global test environment set-up.\n" ) );
}

void uctest_printGlobalTeardown( void ) {
    fprintf_P( test_out, XSTR( "\e[32m[----------]\e[0m Global test environment tear-down.\n" ) );
}

void uctest_printTestRegister( const __flash struct uctest_test *test ) {
    ( void ) test;
}

void uctest_printSuiteStats( const __flash char *suite, int tests ) {
#ifdef __AVR__
    fprintf_P( test_out, XSTR( "\e[32m[----------]\e[0m %d tests from %S" ), tests, suite );
#else
    fprintf_P( test_out, XSTR( "\e[32m[----------]\e[0m %d tests from %s" ), tests, suite );
#endif
}

void uctest_printSuiteStatsBegin( const __flash char *suite, int tests ) {
    uctest_printSuiteStats( suite, tests );
    fprintf( test_out, "\n" );
}

void uctest_printSuiteStatsEnd( const __flash char *suite, int tests, int failed,
                                unsigned long cycles ) {
    ( void ) failed;
    uctest_printSuiteStats( suite, tests );
    fprintf_P( test_out, XSTR( " (%lu cycles)\n" ), cycles );
    fprintf( test_out, "\n" );
}

void uctest_printTest( const __flash char *mode, const __flash struct uctest_test *test ) {
#ifdef __AVR__
    fprintf_P( test_out, XSTR( "\e[32m[ %S ]\e[0m %S.%S" ), mode, test->suite_name,
               test->test_name );
#else
    fprintf_P( test_out, XSTR( "\e[32m[ %s ]\e[0m %s.%s" ), mode, test->suite_name,
               test->test_name );
#endif
}

void uctest_printTestBegin( const __flash struct uctest_test *test ) {
    uctest_printTest( XSTR( "RUN     " ), test );
    fprintf( test_out, "\n" );
}

void uctest_printTestSuccess( const __flash struct uctest_test *test, unsigned long cycles ) {
    uctest_printTest( XSTR( "      OK" ), test );
    fprintf_P( test_out, XSTR( " (%lu cycles)\n" ), cycles );
}

void uctest_printFailedTestsAtEnd( int tests ) {
    fprintf_P( test_out, XSTR( "\n%d FAILED TESTS\n" ), tests );
}

void uctest_printAllTestsEnd( int testsSuccess, int testsFailure ) {
    fprintf_P( test_out, XSTR( "\e[32m[  PASSED  ]\e[0m %d tests.\n" ), testsSuccess );
    if( testsFailure ) {
        fprintf_P( test_out, XSTR( "\e[31m[  FAILED  ]\e[0m %d tests, listed below:\n" ),
                   testsFailure );
    }
}

void uctest_printTestFailure( const __flash struct uctest_test *test, unsigned long cycles ) {
#ifdef __AVR__
    fprintf_P( test_out, XSTR( "\e[31m[  FAILED  ]\e[0m %S.%S" ), test->suite_name,
               test->test_name );
#else
    fprintf_P( test_out, XSTR( "\e[31m[  FAILED  ]\e[0m %s.%s" ), test->suite_name,
               test->test_name );
#endif
    fprintf_P( test_out, XSTR( " (%lu cycles)\n" ), cycles );
}

void uctest_printFailureLocation( const char *file, int line ) {
    fprintf_P( test_out, XSTR( "%s:%d: Failure\n" ), file, line );
}

void uctest_printUserMessage( const __flash char *fmt, va_list args ) {
    vfprintf_P( test_out, fmt, args );
}
#endif

#if !defined( CONFIG_OUTPUT_NONE )
const __flash char uctest_eqFailMsg[] =
    "Expected equality of these values :\n %s\n   Which is : %d\n %s\n   Which is : %d\n";

const __flash char uctest_nearFailMsg[] =
    "The difference between %s and %s is %d, which exceeds %s, where\n  %s "
    "evaluates to %d\n  %s evaluates to %d, and\n  %s evaluates to %d.\n";

const __flash char uctest_simpleFailMsg[] = "Expected: %s, actual: %d vs %d\n";

const __flash char uctest_deathFailMsg[] =
    "Death test: %s\n    Result: failed to die.\n[  DEATH   ]\n";
#endif
